module GCompile where

import Control.Monad
import Control.Monad.State
import Language
import GCode
import LambdaLifter

data GCState = GCState {
    snum :: Int
}

type GCM = State GCState

initstate = GCState 0

genNum :: GCM Int
genNum = do
    s <- get
    put (s { snum = 1 + snum s })
    return (snum s)

genId :: GCM String
genId = genNum >>= (return.('z':).show)

comp_builtin k n = SComb ("builtin." ++ n) k
    (foldr (flip App) (builtin n) [Bound i | i <- [0..k-1]])

builtins = [
    SComb "builtin.pair" 5 (Bound 0 `App` Bound 2 `App` Bound 1),
    SComb "builtin.fst" 3 (Bound 0 `App` builtin "k_2_1"),
    SComb "builtin.snd" 3 (Bound 0 `App` builtin "k_2_2"),
    SComb "builtin.k_2_2" 2 (Bound 0),
    SComb "builtin.k_2_1" 2 (Bound 1),
    SComb "builtin.inl" 5 (Bound 1 `App` Bound 2),
    SComb "builtin.inr" 5 (Bound 0 `App` Bound 2),
    SComb "builtin.case" 4 (Bound 0),
    SComb "builtin.unit" 0 (IntV 0),
    SComb "builtin.Star" 10 (Bound 9),
    SComb "builtin.Unit" 10 (Bound 8),
    SComb "builtin.Bool" 10 (Bound 7),
    SComb "builtin.Int" 10 (Bound 6),
    SComb "builtin.Float" 10 (Bound 5),
    SComb "builtin.Char" 10 (Bound 4),
    SComb "builtin.Pair" 12 (Bound 3 `App` Bound 11 `App` Bound 10),
    SComb "builtin.Sum" 12 (Bound 2 `App` Bound 11 `App` Bound 10),
    SComb "builtin.Arrow" 12 (Bound 1 `App` Bound 11 `App` Bound 10),
    SComb "builtin.Pi" 12 (Bound 0 `App` Bound 11 `App` Bound 10),
    SComb "builtin.casetype" 1 (Bound 0),
    SComb "builtin.true" 0 (BoolV True),
    SComb "builtin.false" 0 (BoolV False),
    SComb "builtin.Fix" 0 (builtin "fix"),
    SComb "builtin.wrap" 3 (Bound 0),
    SComb "builtin.unwrap" 3 (Bound 0),
    SComb "builtin.IO" 0 (IntV 0), -- fsck, IO powinien byc 11tym typem magicznym
    SComb "builtin.return" 3 (builtin "pair" `App` Bound 2 `App` builtin "Unit" `App` Bound 1 `App` Bound 0),
--    SComb "builtin.bind" 5 (Let (Bound 2 `App` Bound 0) (Bound 2 `App` (builtin "fst" `App` Bound 5 `App` builtin "Unit" `App` Bound 0) `App` (builtin "snd" `App` Bound 5 `App` builtin "Unit" `App` Bound 0))),
    comp_builtin 4 "if",
    comp_builtin 2 "fix",
    comp_builtin 2 "addint",
    comp_builtin 2 "subint",
    comp_builtin 2 "mulint",
    comp_builtin 2 "divint",
    comp_builtin 2 "remint",
    comp_builtin 2 "eqint",
    comp_builtin 2 "neint",
    comp_builtin 2 "ltint",
    comp_builtin 2 "leint",
    comp_builtin 1 "negint",
    comp_builtin 2 "addfloat",
    comp_builtin 2 "subfloat",
    comp_builtin 2 "mulfloat",
    comp_builtin 2 "divfloat",
    comp_builtin 2 "remfloat",
    comp_builtin 2 "eqfloat",
    comp_builtin 2 "nefloat",
    comp_builtin 2 "ltfloat",
    comp_builtin 2 "lefloat",
    comp_builtin 1 "negfloat",
    comp_builtin 1 "floattoint",
    comp_builtin 1 "inttofloat",
    comp_builtin 1 "sin",
    comp_builtin 1 "cos",
    comp_builtin 1 "tan",
    comp_builtin 1 "sinh",
    comp_builtin 1 "cosh",
    comp_builtin 1 "tanh",
    comp_builtin 1 "asin",
    comp_builtin 1 "acos",
    comp_builtin 1 "atan",
    comp_builtin 2 "atan2",
    comp_builtin 1 "ceil",
    comp_builtin 1 "floor",
    comp_builtin 1 "round",
    comp_builtin 1 "abs",
    comp_builtin 2 "pow",
    comp_builtin 1 "sqrt",
    comp_builtin 1 "cbrt",
    comp_builtin 1 "log",
    comp_builtin 1 "log10",
    comp_builtin 2 "and",
    comp_builtin 2 "or",
    comp_builtin 1 "not",
    comp_builtin 2 "eqv",
    comp_builtin 2 "neqv"
    ]

precompiled = [
    GFun "builtin.bind" 5 [Push 4,Eval,Push 3,MkAp,Eval,Push 0,PushGlobal "builtin.Unit",Push 3,PushGlobal "builtin.snd",MkAp,MkAp,MkAp,Push 1,PushGlobal "builtin.Unit",Push 4,PushGlobal "builtin.fst",MkAp,MkAp,MkAp,Push 6,MkAp,MkAp,Update 7,Pop 6,Unwind],
    GFun "builtin.readint" 1 [Eval,CallCFun (Just CInt) [] "my_readint",PushGlobal "builtin.Unit",PushGlobal "builtin.Int",PushGlobal "builtin.pair",MkAp,MkAp,MkAp,MkAp,Update 1,Unwind],
    GFun "builtin.readfloat" 1 [Eval,CallCFun (Just CDouble) [] "my_readfloat",PushGlobal "builtin.Unit",PushGlobal "builtin.Int",PushGlobal "builtin.pair",MkAp,MkAp,MkAp,MkAp,Update 1,Unwind],
    GFun "builtin.readchar" 1 [Eval,CallCFun (Just CChar) [] "my_readchar",PushGlobal "builtin.Unit",PushGlobal "builtin.Char",PushGlobal "builtin.pair",MkAp,MkAp,MkAp,MkAp,Update 1,Unwind],
    GFun "builtin.printint" 2 [Push 1,Eval,Pop 1,Eval,CallCFun Nothing [CInt] "my_printint",PushGlobal "builtin.unit",PushGlobal "builtin.Unit",PushGlobal "builtin.Unit",PushGlobal "builtin.pair",MkAp,MkAp,MkAp,MkAp,Update 1,Unwind],
    GFun "builtin.printfloat" 2 [Push 1,Eval,Pop 1,Eval,CallCFun Nothing [CDouble] "my_printfloat",PushGlobal "builtin.unit",PushGlobal "builtin.Unit",PushGlobal "builtin.Unit",PushGlobal "builtin.pair",MkAp,MkAp,MkAp,MkAp,Update 1,Unwind],
    GFun "builtin.printchar" 2 [Push 1,Eval,Pop 1,Eval,CallCFun Nothing [CChar] "my_printchar",PushGlobal "builtin.unit",PushGlobal "builtin.Char",PushGlobal "builtin.Unit",PushGlobal "builtin.pair",MkAp,MkAp,MkAp,MkAp,Update 1,Unwind]
    ]

(+++) = liftM2 (++)
r = return

compile l = evalState compile' initstate where
    compile' = do
        r <- mapM compF (builtins ++ l)
        return (r ++ precompiled)

compF (SComb id n e) = do
    code <- compR e' n
    return (GFun id n (peephole code)) where
        hb k = subst k (Free (Local k))
        e' = foldr ($) e (map hb [0..n-1])

compR (IntV k) d = r [PushInt k, Update (d+1), Pop d, Return]
compR (FloatV k) d = r [PushFloat k, Update (d+1), Pop d, Return]
compR (BoolV k) d = r [PushBool k, Update (d+1), Pop d, Return]
compR (CharV k) d = r [PushChar k, Update (d+1), Pop d, Return]
compR (Let t1 t2) d = compC t1 d +++ compR (unbound d t2) (d+1)
compR (App (App (App (App (Free (Builtin "if")) tc) _) t1) t2) d = do
    lf <- genId
    compE tc d +++ r [JFalse lf] +++ compR t1 d +++
        r [Label lf] +++ compR t2 d
compR (App (App (Free (Builtin n)) t1) t2) d 
    | Just i <- lookup n builtin_binI = 
        compE t2 d +++ compE t1 (d+1) +++ i +++ r [Update (d+1), Pop d, Return]
compR (App (Free (Builtin n)) t) d 
    | Just i <- lookup n builtin_unI = 
        compE t d +++ i +++ r [Update (d+1), Pop d, Return]
compR e d = compC e d +++ r [Update (d+1), Pop d, Unwind]

builtin_binI = [
    ("addint",   r [AddInt]),
    ("subint",   r [SubInt]),
    ("mulint",   r [MulInt]),
    ("divint",   r [DivInt]),
    ("remint",   r [RemInt]),
    ("eqint",    r [EqInt]),
    ("neint",    r [NEInt]),
    ("ltint",    r [LTInt]),
    ("leint",    r [LEInt]),
    ("addfloat", r [AddFloat]),
    ("subfloat", r [SubFloat]),
    ("mulfloat", r [MulFloat]),
    ("divfloat", r [DivFloat]),
    ("remfloat", r [RemFloat]),
    ("eqfloat",  r [EqFloat]),
    ("nefloat",  r [NEFloat]),
    ("ltfloat",  r [LTFloat]),
    ("lefloat",  r [LEFloat]),
    ("and",      r [AndBool]),
    ("or",       r [OrBool]),
    ("eqv",      r [EqvBool]),
    ("neqv",     r [NEvBool]),
    ("atan2",    r [CallCFun (Just CDouble) [CDouble, CDouble] "atan2"]),
    ("pow",      r [CallCFun (Just CDouble) [CDouble, CDouble] "atan2"])
    ]

builtin_unI = [
    ("negint",   r [NegInt]),
    ("negfloat", r [NegFloat]),
    ("not",      r [NotBool]),
    ("inttofloat", r [IntToFloat]),
    ("floattoint", r [FloatToInt]),
    ("sin",      r [CallCFun (Just CDouble) [CDouble] "sin"]),
    ("cos",      r [CallCFun (Just CDouble) [CDouble] "cos"]),
    ("tan",      r [CallCFun (Just CDouble) [CDouble] "tan"]),
    ("sinh",     r [CallCFun (Just CDouble) [CDouble] "sinh"]),
    ("cosh",     r [CallCFun (Just CDouble) [CDouble] "cosh"]),
    ("tanh",     r [CallCFun (Just CDouble) [CDouble] "tanh"]),
    ("asin",     r [CallCFun (Just CDouble) [CDouble] "asin"]),
    ("acos",     r [CallCFun (Just CDouble) [CDouble] "acos"]),
    ("atan",     r [CallCFun (Just CDouble) [CDouble] "atan"]),
    ("ceil",     r [CallCFun (Just CDouble) [CDouble] "ceil"]),
    ("floor",    r [CallCFun (Just CDouble) [CDouble] "floor"]),
    ("round",    r [CallCFun (Just CDouble) [CDouble] "round"]),
    ("abs",      r [CallCFun (Just CDouble) [CDouble] "abs"]),
    ("sqrt",     r [CallCFun (Just CDouble) [CDouble] "sqrt"]),
    ("cbrt",     r [CallCFun (Just CDouble) [CDouble] "cbrt"]),
    ("log",      r [CallCFun (Just CDouble) [CDouble] "log"]),
    ("log10",    r [CallCFun (Just CDouble) [CDouble] "log10"])
    ]

compC (Free (Global i)) _ = r [PushGlobal i]
compC (Free (Builtin i)) _ = r [PushGlobal ("builtin." ++ i)]
compC (Free (Local k)) d = r [Push (d-k-1)]
compC (App (App (Free (Builtin "fix")) _) t) d =
    r [Alloc 1, Push 0] +++ compC t (d+2) +++ r [MkAp, Update 1]
compC (App t1 t2) d = compC t2 d +++ compC t1 (d+1) +++ r [MkAp]
compC (Let t1 t2) d = compC t1 d +++ compC (unbound d t2) (d+1) +++ r [Slide 1]
compC (IntV v) d = r [PushInt v]
compC (FloatV v) d = r [PushFloat v]
compC (BoolV v) d = r [PushBool v]
compC (CharV v) d = r [PushChar v]
compC t d = error (show t)

compE (Let t1 t2) d = compC t1 d +++ compE (unbound d t2) (d+1) +++ r [Slide 1]
compE (IntV v) d = r [PushInt v]
compE (FloatV v) d = r [PushFloat v]
compE (BoolV v) d = r [PushBool v]
compE (CharV v) d = r [PushChar v]
compE (App (App (App (App (Free (Builtin "if")) tc) _) t1) t2) d = do
    [lf,la] <- replicateM 2 genId
    compE tc d +++ r [JFalse lf] +++ compE t1 d +++ r [Jump la] +++
        r [Label lf] +++ compE t2 d +++ r [Label la]
compE (App (App (Free (Builtin n)) t1) t2) d 
    | Just i <- lookup n builtin_binI = 
        compE t2 d +++ compE t1 (d+1) +++ i
compE (App (Free (Builtin n)) t) d 
    | Just i <- lookup n builtin_unI = compE t d +++ i
compE t d = compC t d +++ r [Eval]

peephole (Alloc i:Alloc j:t) = peephole (Alloc (i+j):t)
peephole (Slide i:Slide j:t) = peephole (Slide (i+j):t)
peephole (Push _:Pop k:t) | k > 0 = peephole (Pop (k-1):t)
peephole (PushGlobal _:Pop k:t) | k > 0 = peephole (Pop (k-1):t)
peephole (PushInt _:Pop k:t) | k > 0 = peephole (Pop (k-1):t)
peephole (PushFloat _:Pop k:t) | k > 0 = peephole (Pop (k-1):t)
peephole (PushBool _:Pop k:t) | k > 0 = peephole (Pop (k-1):t)
peephole (PushChar _:Pop k:t) | k > 0 = peephole (Pop (k-1):t)
peephole (Pop 0:t) = peephole t
peephole (Eval:Eval:t) = peephole (Eval:t)
peephole (h:t) = h:peephole t
peephole [] = []


