module LambdaLifter(SComb(..), lambdalift) where

import Control.Monad.State
import Control.Monad
import qualified Data.Set as Set
import Data.Set(Set)
import Language

data SComb = SComb String Int Term deriving (Eq, Show)

data LLState = LLState {
    scname :: String,
    snum :: Int,
    output :: [SComb]
}

type LLM = State LLState

initstate = LLState "prog" 0 []

genNum :: LLM Int
genNum = do
    s <- get
    put (s { snum = 1 + snum s })
    return (snum s)

genName :: LLM String
genName = do
    n <- genNum
    s <- get
    return (scname s ++ show n)

changeName :: String -> LLM ()
changeName n = do
    s <- get
    put (s { scname = n })

outComb :: SComb -> LLM ()
outComb c = do
    s <- get
    put (s { output = c:output s })

llift :: Int -> Term -> LLM Term
llift d t@(Lam _) = do
    let (t', k) = dropLambdas d t
    t'' <- llift (d+k) t'
    let fv = Set.toList $ Set.filter (localbelow d) (freevars t'')
    let av = reverse ([v | Local v <- fv] ++ [d..d+k-1])
    let ct = ssubst (hsubst (zip av [0..])) 0 t''
    n <- genName
    outComb (SComb n (k+length fv) ct)
    return (foldl (\x y -> App x (Free y)) (Free (Global n)) fv)
llift d (App t1 t2) = liftM2 App (llift d t1) (llift d t2)
llift d (Let t1 t2) = 
    liftM2 Let (llift d t1) (liftM (bound d) (llift (d+1) (unbound d t2)))
llift d (Ann t _) = llift d t
llift d (CaseType v t1 t2 t3 t4 t5 t6 t7 t8 t9 t10) =
    llift d (builtin "casetype" `App` v `App` t1 `App` t2 `App` t3 `App` t4
        `App` t5 `App` t6 `App` t7 `App` t8 `App` t9 `App` t10)
llift d Star = llift d (builtin "Star")
llift d (Pi t1 t2) = do
    if Local d `Set.member` freevars (unbound d t2)
        then llift d (builtin "Pi" `App` t1 `App` (Lam t2)) 
        else llift d (builtin "Arrow" `App` t1 `App` t2)
llift d t = return t

hsubst :: [(Int, Int)] -> Int -> Term -> Term
hsubst m d (Free (Local i)) = case lookup i m of Just k -> Bound (d+k)
hsubst _ _ t = t

lliftall :: Term -> LLM [SComb]
lliftall t = do
    t' <- llift 0 t
    outComb (SComb "prog" 0 (builtin "snd" `App` builtin "Int" `App` builtin "Unit" `App` (t' `App` builtin "unit")))
    s <- get
    return (output s)

dropLambdas :: Int -> Term -> (Term, Int)
dropLambdas d (Lam t) = (t', k+1) where (t', k) = dropLambdas (d+1) (unbound d t)
dropLambdas d t = (t, 0)

localbelow :: Int -> Id -> Bool
localbelow k (Local d) = d < k
localbelow _ _ = False

lambdalift :: Term -> [SComb]
lambdalift t = evalState (lliftall t) initstate

