module Preprocessor(preprocess) where

import Language
import Data.Map(Map)
import qualified Data.Map as Map
import qualified Data.Set as Set
import Data.Graph.Inductive(Graph)
import qualified Data.Graph.Inductive as G
import Data.List(nub)
import Data.Maybe(fromJust,isJust)

prep d e (App t1 t2) = App (prep d e t1) (prep d e t2)
prep d e (AppTy t1 t2) = AppTy (prep d e t1) (prep d e t2)
prep d e (Ann t1 t2) = Ann (prep d e t1) (prep d e t2)
prep d e (Star) = Star
prep d e (IntV v) = IntV v
prep d e (FloatV v) = FloatV v
prep d e (BoolV v) = BoolV v
prep d e (CharV v) = CharV v
prep d e (CaseType t0 t1 t2 t3 t4 t5 t6 t7 t8 t9 t10) = 
    let [pt0,pt1,pt2,pt3,pt4,pt5,pt6,pt7,pt8,pt9,pt10] = map (prep d e) [t0,t1,t2,t3,t4,t5,t6,t7,t8,t9,t10]
    in (CaseType pt0 pt1 pt2 pt3 pt4 pt5 pt6 pt7 pt8 pt9 pt10)
prep d e (Free (Global i)) | Just di <- Map.lookup i e = 
    boundAll d (prep d e di)
prep d e (Free i) = Free i
prep d e (PrimLam s t) = Lam (prep (d+1) (Map.insert s (local d) e) t)
prep d e (PrimPi s t1 t2) = 
    Pi (prep d e t1) (prep (d+1) (Map.insert s (local d) e) t2)
prep d e (PrimLet de t) = prepLet d e (orderDecls de) t
prep d e t = error (show t)

prepLet d e [] t = prep d e t
prepLet d e ([(i,t1)]:r) t2 | (Global i) `Set.notMember` freevars t1 =
    Let (prep d e t1) (prepLet (d+1) (Map.insert i (local d) e) r t2)
prepLet d e (de:r) t2 = Let de' (prepLet (d+1) e' r t2) where
    de' = builtin "fix" `App` (prep d e t') `App` (Lam (prep (d+1) e' v'))
    (e', v', t') = hEnv id de
    hEnv hf [(i,Ann v t)] = (e', v, t) where
        e' = Map.insert i (local d) e
    hEnv hf [(i1,Ann v1 t1),(i2,Ann v2 t2)] = (e', v', t') where
        e' = Map.insert i1 (builtin "fst" `App` t1 `App` t2 `App` hf (local d)) $
             Map.insert i2 (builtin "snd" `App` t1 `App` t2 `App` hf (local d)) e
        v' = builtin "pair" `App` t1 `App` t2 `App` v1 `App` v2
        t' = builtin "Pair" `App` t1 `App` t2
    hEnv hf ((i,Ann v t):r) = (e'', v'', t'') where
        (e', v', t') = hEnv hf' r
        hf' = hf . (\x -> builtin "fst" `App` t' `App` t `App` x)
        e'' = Map.insert i (builtin "snd" `App` t' `App` t `App` hf (local d)) e'
        v'' = builtin "pair" `App` t' `App` t `App` v' `App` v
        t'' = builtin "Pair" `App` t' `App` t

preprocess = prep 0 Map.empty

orderDecls :: [(String,Term)] -> [[(String,Term)]]
orderDecls dl = G.topsort' compGraph where
    justLookup k = fromJust . Map.lookup k
    depGraph :: G.Gr (String,Term) ()
    depGraph = G.mkGraph nl el where
        nl = zip [1..] dl
        numb = Map.fromList (map (\(a,(s,_)) -> (s,a)) nl)
        el = foldl f [] dl
        f el (s,e) = Set.fold g el (freevars e) where
            sv = justLookup s numb
            g (Global v) l = case Map.lookup v numb of
                Just ev -> (ev,sv,()):l
                Nothing -> l
            g _ l = l
    comp = G.scc depGraph
    compGraph :: G.Gr [(String,Term)] ()
    compGraph = G.mkGraph nl el where
        nl' = zip [1..] comp
        nl = map (\(n,l) -> (n,map (fromJust . G.lab depGraph) l)) nl'
        numb = Map.fromList 
                (foldl (\l (n,m) -> foldl (\l x -> (x,n):l) l m) [] nl')
        el = nub (map (\(s,d,l) -> (justLookup s numb,justLookup d numb,l)) 
                                        (G.labEdges depGraph))

